import logging


class Validator:
    def __init__(self, value, value_name):
        self.value = value
        self.value_name = value_name

    def check_is_not_instance(self, instance_type):

        if isinstance(self.value, instance_type):
            raise TypeError(
                f"The {self.value_name} must be of type {str(instance_type)}, you supplied '{str(type(self.value))}' instead."
            )
        return self

    def check_allowed_value(self, allowed_values: list):
        """
        Checks that value exists in the provided list

        Parameters
        ----------
        allowed_values : list
            The list of allowed values

        Returns
        -------
        Validator
        """

        if self.value not in allowed_values:
            raise ValueError(
                f"The {self.value_name} must be one of {str(allowed_values)}, you supplied '{self.value}' instead."
            )
        return self

    def make_singular(self):
        """
        Makes a plural string singular

        Returns
        -------
        Validator
        """

        if isinstance(self.value, str):
            self.value = self.value.rstrip("s")
            logging.debug(
                f"The value of {self.value_name} has had any 's' characters stripped from the right to make it singular"
            )
        return self

    def make_lower(self):
        """
        Makes a string lowercase

        Returns
        -------
        Validator
        """

        if isinstance(self.value, str):
            self.value = self.value.lower()
            logging.debug(f"The value of {self.value_name} has been made lower case")
        return self

    def set_default_value_if_none(self, default):
        """
        Sets a default value if the current value is None

        Parameters
        ----------
        default
            The default value

        Returns
        -------
        Validator
        """

        if self.value is None:
            self.value = default
            logging.debug(
                f"The value of {self.value_name} has been updated from None to {default}"
            )
        return self

    def override_value(self, override_flag: bool, override_value):
        """
        Overrides the current value of the ovverride_flag is True

        Parameters
        ----------
        override_flag : bool
            Whether or not to override the current value
        override_value
            The value to use to override the existing values

        Returns
        -------
        Validator
        """

        if override_flag:
            self.value = override_value
            logging.debug(
                f"The value of {self.value_name} has been overriden with {override_value}"
            )
        return self

    def discard_dict_keys_none_value(self):
        """
        Discards dictionary key, value pairs where the value is None

        Returns
        -------
        Validator
        """

        if isinstance(self.value, dict):
            invalid_keys = [key for key, value in self.value.items() if value is None]
            if len(invalid_keys) > 0:
                logging.info(
                    f"The values for the keys {str(invalid_keys)} are None and have thus been removed from {self.value_name}"
                )
                for key in invalid_keys:
                    self.value.pop(key, None)
        return self

    def get_dict_values(self):
        """
        Gets a list of values from a dictionary

        Returns
        -------
        Validator
        """

        if isinstance(self.value, dict):
            self.value = list(self.value.values())
        return self

    def filter_list_using_first_character(self, first_character: str):
        """
        Filters a list of strings by looking to see if the first character matches the provided

        Parameters
        ----------
        first_character : str
            The character to look for in the first character of each element

        Returns
        -------
        Validator
        """

        if isinstance(self.value, list):
            to_remove = []
            for value in self.value:
                if isinstance(value, str) and (
                    len(value) > 0 and value[0] == first_character
                ):
                    to_remove.append(value)
            self.value = [value for value in self.value if value not in to_remove]
        return self

    def check_subset_of_list(self, superset: list, superset_name: str):
        """
        Checks if one list is a subset of another

        Parameters
        ----------
        superset : list
            The superset to check of the value is a subset of
        superset_name : list
            The name of the superset

        Returns
        -------
        Validator
        """

        if isinstance(self.value, list):
            if len(set(self.value) - set(superset)) > 0:
                raise ValueError(
                    f"""The values {str(set(self.value) - set(superset))} exist in the {self.value_name}
                                   but do not exist in the {superset_name}."""
                )
        return self

    def check_no_intersection_with_list(self, other_list: list, list_name: str):
        """
        Checks that the value has no intersection with a provided list

        Parameters
        ----------
        other_list : list
            The list to check the value has no intersection with
        list_name : list
            The name of the list

        Returns
        -------
        Validator

        """

        if isinstance(self.value, list):
            if len(set(other_list).intersection(set(self.value))) > 0:
                err = f"""The columns {str(set(other_list).intersection(set(self.value)))} are specified in {self.value_name}
                         yet they contain null (NaN) values for some rows in the provided data. Null values are not 
                         allowed for required fields. Please ensure that required columns do not contain ANY null 
                         values or specify a default value in your mapping by specifying a dictionary with the keys
                         "column" and "default"."""
                logging.error(err)
                raise ValueError(err)

    def check_entries_are_strings_or_dict_containing_key(self, expected_key):
        missing_items = []

        for item in self.value:
            if not (isinstance(item, dict) or isinstance(item, str)):
                missing_items.append(f"{item} is not a string or dictionary")
                continue

            if isinstance(item, dict):
                if expected_key not in item:
                    missing_items.append(
                        f"{item} does not contain the mandatory 'source' key"
                    )
                    continue
                if not isinstance(item[expected_key], str):
                    missing_items.append(
                        f"{item[expected_key]} in {item} is not a string"
                    )
                    continue

        if len(missing_items) > 0:
            raise ValueError(
                f"The value {self.value} provided in {self.value_name} is invalid. {', '.join(missing_items)}."
            )

        return self



import logging
import lusid
import pathlib
from lusidtools.lpt import stdargs
from lusidtools.logger import LusidLogger


def parse(extend=None, args=None):
    return (
        stdargs.Parser(
            "Flush Transactions", ["scope", "optional_portfolio", "date_range", "group", "flush_scope"],
        )
        .extend(extend)
        .parse(args)
    )


def transaction_batcher_by_character_count(
    scope: str, code: str, host: str, input_lst: list, maxCharacterCount: int = 4000
):
    """
    Takes a given input list of transactions or transactionIDs and batches them based on a maximum number of
    characters allowed in the URI

    Parameters
    ----------
    scope : str
        Scope of the portfolio in question
    code : str
        Code of the portfolio in question
    host : str
        Host taken from the configuration of the api client. Represents the start the base url involved in the calls
    input_lst : list
        A list containing all the transaction information
    maxCharacterCount : int
        Maximum number of characters allowed in the URI

    Returns
    -------
    batched_data : list of lists
        Returns a list with each item being a batch (which is itself, a list of transactions)

    """

    # Determine how many characters are fixed to be included in url before the addition of query params
    base_url = f"{host}/api/transactionportfolios/{scope}/{code}/transactions?"
    unavoidable_character_count = len(base_url)
    remaining_count = maxCharacterCount - unavoidable_character_count

    # Build the batches
    batches = []
    curr_count = 0
    curr_batch = []
    for txn_id in input_lst:
        if curr_count + len(f"transactionIds={txn_id}&") >= remaining_count:
            batches.append(curr_batch)
            curr_batch = []
            curr_count = 0

        curr_batch.append(txn_id)
        curr_count = curr_count + len(f"transactionIds={txn_id}&")

    # Create remainder final batch if any remainder exists
    if curr_batch:
        batches.append(curr_batch)

    logging.info(
        f"Batched the transactions into {len(batches)} batches based on the character limit of {maxCharacterCount}"
    )

    return batches


class TxnGetter:
    def __init__(
        self, scope, portfolio, start_date, end_date, transaction_portfolios_api
    ):
        self.scope = scope
        self.portfolio = portfolio
        self.start_date = start_date
        self.end_date = end_date
        self.transaction_portfolios_api = transaction_portfolios_api
        self.stop = False

    def _get_transactions(self, scope, portfolio, start_date, end_date, page=None):
        # make API call to LUSID
        if page is None:
            return self.transaction_portfolios_api.get_transactions(
                scope,
                portfolio,
                from_transaction_date=start_date,
                to_transaction_date=end_date,
                limit=5000,
            )
        else:
            return self.transaction_portfolios_api.get_transactions(
                scope,
                portfolio,
                from_transaction_date=start_date,
                to_transaction_date=end_date,
                limit=5000,
                page=page,
            )

    def __iter__(self):
        # get first page and get next page token
        self.txns = self._get_transactions(
            self.scope, self.portfolio, self.start_date, self.end_date
        )
        self.next_page = self.txns.next_page
        return self

    def __next__(self):
        # if there is a next page token get the next page
        if self.stop:
            raise StopIteration

        result = self.txns
        if self.next_page:
            self.txns = self._get_transactions(
                self.scope,
                self.portfolio,
                self.start_date,
                self.end_date,
                self.next_page,
            )
            self.next_page = self.txns.next_page
            return result
        else:
            self.stop = True
            return result


def get_paginated_txns(
    scope, portfolio, start_date, end_date, transaction_portfolios_api
):
    """
    Gets paginated transactions in a given time window

    Parameters
    ----------
    scope : str
        Scope of the portfolio in question
    portfolio : str
        Code of the portfolio in question
    start_date : str
        Lower bound from when transactions will be looked for
    end_date : str
        Upper bound to when transactions will be looked for
    transaction_portfolios_api : TransactionPortfoliosApi
        The instance of the API used for the requests

    Returns
    -------
    List of all responses from LUSID, each response relating to a page of transactions

    """
    txn_getter = TxnGetter(
        scope, portfolio, start_date, end_date, transaction_portfolios_api
    )
    return [response for response in txn_getter]


def get_portfolios_from_group(expanded_group):
    """
    Recursive function to extract all the portfolio information from the get_portfolio_group_expansion endpoint

    Parameters
    ----------
    expanded_group : ExpandedGroup
        Response from get_portfolio_group_expansion endpoint containing all the portfolio scopes and codes in this group
        as well as further ExpandedGroup objects for each of the sub-groups

    Returns
    -------
    A list of tuples with each item describing a portfolios in the form (scope, code)
    """
    temp_portfolio_lst = [
        (portfolio.id.scope, portfolio.id.code) for portfolio in expanded_group.values
    ]

    if expanded_group.sub_groups:
        for sub_group in expanded_group.sub_groups:
            temp_portfolio_lst.extend(get_portfolios_from_group(sub_group))

    return temp_portfolio_lst


def get_all_transactions_from_portfolio_list(args, portfolio_list, transaction_portfolios_api):
    """
    Gets all transactions from the list of portfolio scope and codes that are passed in. Pagination is handled.

    Parameters
    ----------
    args : Namespace
            The arguments taken in when command is run
    portfolio_list : [tuple(str,str)]
            A list of tuples containing portfolio scope and codes
    transaction_portfolios_api : lusid.TransactionPortfoliosApi
            An instantiated TransactionPortfoliosApi

    Returns
    -------
    txn_dict : dict
        Dictionary where the keys are tuples of portfolio scope and codes and the values are a list of transactions

    """
    # Get transactions from these portfolios
    return {
        (portfolio_scope, portfolio_code): [
            transaction
            for page
            in get_paginated_txns(
                portfolio_scope,
                portfolio_code,
                args.start_date,
                args.end_date,
                transaction_portfolios_api,
            )
            for transaction
            in page.values]
        for (portfolio_scope, portfolio_code)
        in portfolio_list
    }


def get_all_txns(args):
    """
    Gets all the transactions in a given time window

    Parameters
    ----------
    args : Namespace
            The arguments taken in when command is run

    Returns
    -------
    Dictionary with key: (scope, code) and value: list of transactions
    """

    api_factory = lusid.utilities.ApiClientFactory(api_secrets_filename=args.secrets)
    transaction_portfolios_api = api_factory.build(lusid.api.TransactionPortfoliosApi)

    if args.group:
        # Get a list of all portfolios from the group
        groups_api = api_factory.build(lusid.api.PortfolioGroupsApi)

        portfolios_response = groups_api.get_portfolio_group_expansion(
            scope=args.scope, code=args.portfolio
        )
        portfolio_lst = get_portfolios_from_group(portfolios_response)

        return get_all_transactions_from_portfolio_list(args, portfolio_lst, transaction_portfolios_api)

    elif args.flush_scope:
        # Get a list of all portfolios in the scope
        portfolios_api = api_factory.build(lusid.api.PortfoliosApi)

        portfolio_lst = [
            (portfolio.id.scope, portfolio.id.code) for
            portfolio in
            portfolios_api.list_portfolios_for_scope(scope=args.scope).values
        ]

        return get_all_transactions_from_portfolio_list(args, portfolio_lst, transaction_portfolios_api)

    else:
        # Return the portfolio and scope passed in
        return {
            (args.scope, args.portfolio): [
                transaction
                for page in get_paginated_txns(
                    args.scope,
                    args.portfolio,
                    args.start_date,
                    args.end_date,
                    transaction_portfolios_api,
                )
                for transaction in page.values
            ]
        }


def flush(args):
    """
        Cancels all transactions found in a given date range for a portfolio, portfolio group or scope

        Parameters
        ----------
        args : Namespace
            The arguments taken in when command is run

        Returns
        ----------
        Number of successful batches : int
        Number of failed batches : int
    """

    # Initialise the api
    api_factory = lusid.utilities.ApiClientFactory(api_secrets_filename=args.secrets)
    transaction_portfolios_api = api_factory.build(lusid.api.TransactionPortfoliosApi)

    # Get the Transaction Data
    total_batch_count = 0
    total_failed_batch_count = 0

    for portfolio_id, transactions in get_all_txns(args).items():
        txn_ids = [txn.transaction_id for txn in transactions]

        logging.info(
            f"Looking at portfolio with scope: {portfolio_id[0]} and code: {portfolio_id[1]}"
        )

        if not txn_ids:
            logging.error("No transactions found between the specified dates")
            continue

        logging.info(
            f"Found {len(txn_ids)} transactions in date range {args.end_date} {args.start_date} "
        )

        # Batch the transactions - Necessary to limit the length of the URI in the API call
        txn_ids_batches = transaction_batcher_by_character_count(
            args.scope,
            args.portfolio,
            api_factory.api_client.configuration.host,
            txn_ids,
        )

        # Flush the transactions
        logging.info("Flushing the transactions")
        local_failure_count = 0
        for batch in txn_ids_batches:
            try:
                transaction_portfolios_api.cancel_transactions(
                    portfolio_id[0], portfolio_id[1], transaction_ids=batch
                )
            except lusid.exceptions.ApiException as e:
                logging.error(
                    f"Batch {txn_ids_batches.index(batch)} in portfolio with scope: {portfolio_id[0]} "
                    f"and code: {portfolio_id[1]} has failed with exception {e} "
                )
                local_failure_count = local_failure_count + 1

        total_failed_batch_count = total_failed_batch_count + local_failure_count
        total_batch_count = total_batch_count + len(txn_ids_batches)

        logging.info(
            f"Done flushing this portfolio with {len(txn_ids_batches) - local_failure_count} successes and {local_failure_count} failures"
        )

    return (total_batch_count - total_failed_batch_count), total_failed_batch_count


def main():
    args = parse()
    if args.debug:
        LusidLogger(args.debug)
    successful_batch_count, failed_batch_count = flush(args)

    if failed_batch_count == 0:
        logging.info("All transaction batches were successfully flushed")
    else:
        logging.info(
            f"The following number of batches were successful: {successful_batch_count}"
        )
        logging.error(
            f"The following number of batches failed to flush: {failed_batch_count}"
        )

    return 0


if __name__ == "__main__":
    main()
